/*---------------------------------------------------------------------------*
 *                                   IT++			             *
 *---------------------------------------------------------------------------*
 * Copyright (c) 1995-2005 by Tony Ottosson, Thomas Eriksson, Pl Frenger,   *
 * Tobias Ringstrm, and Jonas Samuelsson.                                   *
 *                                                                           *
 * Permission to use, copy, modify, and distribute this software and its     *
 * documentation under the terms of the GNU General Public License is hereby *
 * granted. No representations are made about the suitability of this        *
 * software for any purpose. It is provided "as is" without expressed or     *
 * implied warranty. See the GNU General Public License for more details.    *
 *---------------------------------------------------------------------------*/

/*!
  \file
  \brief Selective Repeat ARQ classes
  \author Anders Persson and Tony Ottosson

  1.1

  2005/02/28 15:43:00 
*/

#ifndef __selective_repeat_h
#define __selective_repeat_h

#include <stdlib.h>
#include "protocol/signals_slots.h"
#include "protocol/packet.h"
#include "protocol/front_drop_queue.h"
#include "base/array.h"
#include "base/vec.h"

namespace itpp {

  /*!

  */
class Selective_Repeat_ARQ_Sender {
public:
  Selective_Repeat_ARQ_Sender();

  Selective_Repeat_ARQ_Sender(const int Seq_no_size, const int Buffer_size_factor, const int Link_packet_size, const Ttype Time_out);

  ~Selective_Repeat_ARQ_Sender();

  void set_parameters(const int Seq_no_size,        // # bits in sequence no. 
		      const int Buffer_size_factor, // Link-packet buffer size = 2^(Seq_no_size)*Buffer_size_factor.
		      const int Link_packet_size,    // Size of the link packets in bytes.
		      const Ttype Time_out);        // Idle time before retransmission.

  // -- Slots -- //
  Slot<Selective_Repeat_ARQ_Sender, Packet*> packet_input;            // Receives incoming packets.
  Slot<Selective_Repeat_ARQ_Sender, Array<Packet*> > ack_input;                  // Receives incoming ack/nacks.
  Slot<Selective_Repeat_ARQ_Sender, void*> query_nof_ready_packets;   // Receives incoming query for number of packets ready to transmit.
  Slot<Selective_Repeat_ARQ_Sender, int> packet_output_request;       // Receives incoming packet output requests.

  // -- Signals -- //
  Signal<Array<Packet*> > packet_output;                // Delivers transmitted packets.
  Signal<int> nof_ready_packets;                // Delivers no ready packets.
  Signal<int> buffer_overflow;                  // Signals buffer overflows.
  
  int buffer_size();
  int link_packets_buffered();
  int nof_ready_link_packets();
  int link_packets_queued_waiting_for_transmission();
  Ttype link_packets_max_queuing_time();
  void get_link_packets(const int K, Array<Packet*> &pa);

private:
  void handle_ack_input(Array<Packet*> packet_array); // Take care of incomming ack/nacks.
  void handle_packet_input(Packet *P);          // Take care of incomming packets.
  void handle_packet_output_request(int K);     // Take care of incomming packet requests.
  void handle_query_nof_ready_packets(void*);   // Take care of incomming query for number of packets ready to transmit.
  void retransmit(int Sequence_number);    // Take care of incomming query for number of packets ready to transmit.
  void remove(const int Sequence_number);
  void push_packet_on_tx_buffer(Packet *packet);
  int buffered_non_outstanding();
  int free_sequence_numbers();
  int sequence_number_2_buffer_index(const int Sequence_number);
  void schedule_output(const int Buffer_index, const int Sequence_number, const bool Retransmission);
  void cancel_output(const int Sequence_number);
  void fill_output();
  int feasable_blocks();
  bool parameters_ok;
  Front_Drop_Queue ip_pkt_queue;
  Array<Link_Packet*> input_buffer;
  int input_buffer_size;
  int input_next;
  int input_free_space;  
  int seq_no_size;
  int seq_no;
  int seq_no_max;
  int tx_next;
  int tx_last;
  int outstanding;
  int id;
  Ttype time_out;
  Array<ATimer<Selective_Repeat_ARQ_Sender, int> > timer;
  ivec output_indexes;
  ivec retransmission_indexes;
  int rd_pos;
  int rt_pos;
  int scheduled_total;
  int scheduled_retransmissions;
  int no_retransmit;
  int link_packet_size;
};


/*!

*/
class Selective_Repeat_ARQ_Receiver {
public:
  Selective_Repeat_ARQ_Receiver();

  Selective_Repeat_ARQ_Receiver(const int Seq_no_size);

  ~Selective_Repeat_ARQ_Receiver();

  // -- Slots -- //
  Slot<Selective_Repeat_ARQ_Receiver, Array<Packet*> > packet_input;    // Receives incoming packets.

  // -- Signals -- //
  Signal<Array<Packet*> > ack_output;       // Delivers ack.  
  Signal<Packet*> packet_output; // Delivers received packets.  

  void set_parameters(const int Seq_no_size); // # bits in sequence no. 

private:
  bool greater_modulo_L(const int a, const int b);
  void handle_packet_input(Array<Packet*>); // Take care of incomming packets.
  int seq_no_size;
  int seq_no_max;
  Array<Link_Packet*> rx_buffer;
  int Rnext;
  int id;
  bool parameters_ok;
};


} // namespace itpp

#endif //__selective_repeat_h

